# File.pm (HCT::File)
#
# Copyright 2008, 2009
# Slade Maurer
#
# This file is part of HCT, the Hardware Complexity Tool
#
# HCT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# HCT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with HCT.  If not, see <http://www.gnu.org/licenses/>.

=head1 NAME

HCT::File - Set of functions for work files and directories.

=head1 DESCRIPTION

This package containts various subroutines to provide work 
with files and directories.

=cut

package HCT::File;

our $VERSION = '0.1.2';

use strict;
use warnings;

require Exporter;
our @ISA    = qw(Exporter);
our @EXPORT = qw(
	scan_path

	is_file
	is_dir
	is_tmp
	is_sys

	get_basename
	get_dirname
	get_extension
);

use HCT::Std;
use HCT::Constants;

use POSIX;
use File::Basename;

=head1 FUNCTIONS
=cut

=over 4

=cut

=item B<scan_path (PATH)>

The PATH is a list of pathes that should be processed:

	scan_path ($path);

As the result will be returned a hash-array, where keys - directories, and
values - list of files from this directories.

	$res = scan_path ($path);

=cut

sub scan_path
{
	my @path = EMPTY_LIST;
	my $result = { };
	
	if (ref ($_[0]) eq 'ARRAY') { @path = @{$_[0]} }
	else { push @path, $_[0] }

	# If nothing to process, then return empty list
	if (scalar (@path) == 0) {
		return $result;
	}

	PATH: while (my $path = shift @path) {

		$path = File::Spec->canonpath(
			($path =~ /^\//) ? $path : join('/' => ($HCT::START_DIR, $path))
		);

		if (!-e $path) {
			println "$path cannot be found." if $HCT::VERBOSE;
			next PATH;
		}

		# if we have a file
		if (is_file ($path)) {
			my $dir = get_dirname ($path);
			$result->{$dir} = [ ] unless defined $result->{$dir};
			push @{$result->{$dir}}, $path;
		}

		# if we have a directory then process it...
		elsif (is_dir ($path)) {
			my $dir = File::Spec->canonpath ($path);

			unless (defined $result->{$dir}) {
				$result->{$dir} = [ ];
			}

			unless (opendir (DIR, $dir)) {
				println "Can not open directory '$dir': $!"	if $HCT::VERBOSE;
				next PATH;
			}

			my @items = grep {
				$_ ne '.' && $_ ne '..'
			} readdir DIR;

			ITEM: foreach (@items) {
				my $item = File::Spec->canonpath (join ('/' => ($dir, $_)));

				next ITEM 
					if (!$HCT::LIST_ALL
						&& $HCT::LIST_TMP == FALSE
						&& is_tmp ($item));

				next ITEM
					if (!$HCT::LIST_ALL
						&& $HCT::LIST_SYS == FALSE
						&& is_sys ($item));

				# if we have a new file then push it to our task list
				if (is_file ($item)) {
					push @{$result->{$dir}}, $item;
				}

				# else if it is a new directory, then we process it
				elsif (is_dir ($item)) {
					if ($HCT::LIST_RECURSIVELY == TRUE) {
						push @path, $item;
					}
				}
			}

			closedir (DIR);
		}
	}

	return $result;
}

=item B<is_tmp (PATH)>

Portable method to test if path is a temporary file 
or directory. Returns true or false.

=cut

sub is_tmp ($)
{
	my $path = shift;

	if (!ON_WINDOWS && get_basename ($path) =~ /\~$/) {
			return TRUE;
	}
		
	return FALSE;
}

=item B<is_sys (PATH)>

Portable method to test if path is a system file 
or directory. Returns true or false.

=cut

sub is_sys ($)
{
	my $path = shift;

	if (!ON_WINDOWS && get_basename ($path) =~ /^\./) {	
			return TRUE;
	}

	return FALSE;
} 

=item B<is_file (PATH)>

Portable method to test if path is a file. The reason: C<-f> doesn't 
work in ActiveState Perl on Windows. Returns true or false.

=cut

sub is_file
{ 
    no strict 'subs';

    my $item = shift @_;

	if (ON_WINDOWS) {
		my $mode = (stat $item)[2];
           $mode = 0 unless $mode;
        if ($mode & 0100000) { return TRUE; } 
        else { return FALSE; }
    } else {
        return (-f $item);  # works on Unix, Linux, CygWin, z/OS
    }
}

=item B<is_dir (PATH)>

Portable method to test if path is a directory. The reason: C<-d> doesn't 
work in ActiveState Perl on Windows. Returns true or false.

=cut

sub is_dir
{
	no strict 'subs';

	my $path = shift @_;

	if (ON_WINDOWS ) {
		my $mode = (stat $path)[2];
		$mode = 0 unless $mode;
		if ($mode & 0040000) { return TRUE } 
        else { return FALSE }
    } else {
		return (-d $path);  # works on Unix, Linux, CygWin, z/OS
	}
}

=item B<get_basename (PATH)>
 
Returns the file name portion of a path by using C<File::Basename> 
subroutines.

=cut

# FIXME: try to use fileparse() insead of basename.

sub get_basename ($)
{
	my $path = shift;
	return basename ($path);
}

=item B<get_dirname (PATH)>

Returns the dir name portion of a path by using C<File::Basename> 
subroutines.

=cut

# FIXME: try to use fileparse() insead of basename.

sub get_dirname ($)
{
	return dirname ($_[0]); 
}

=item B<get_extension (PATH)>

Returns the file extension portion of a path by using C<File::Basename> 
subroutines.

=cut

sub get_extension ($)
{
	my (undef, undef, $extension) = fileparse (shift, qr/([^.]*)$/);
	return $extension;
}

=back

=cut

1;

