# LangDriver.pm (HCT::Lang::LangDriver) - Base object class for the HCT language.
#
# Copyright 2008, 2009
# Slade Maurer
#
# This file is part of HCT, the Hardware Complexity Tool
#
# HCT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# HCT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with HCT.  If not, see <http://www.gnu.org/licenses/>.

=head1 NAME

HCT::Lang::LangDriver - Base class of language used by HCT.

=cut

package HCT::Lang::LangDriver;

our $VERSION = '0.2.0';

use Time::HiRes;

use HCT::File;
use HCT::Std;
use HCT::Constants;
use HCT::Output;

use strict;
use warnings;

=head1 CONSTRUCTOR

=over 4

=item B<new ()>

Creates a new C<HCT::Lang::LangDriver> object.

=back

=cut

sub new
{
	my $inv = shift;
	my $class = ref($inv) || $inv;
	my $self = {
		
		parser	=> undef,
		lexer	=> undef,
		output	=> new HCT::Output,
		metrics => [ [ '', { } ] ],
	};
	bless $self, $class;
	return $self;
}

sub clone
{
	my $self = shift;
	my $clone = $self->new (@_);
	return $clone;
}

sub test
{
	my $self = shift;
	my $error = FALSE;
	
	unless ( defined $self->{parser} ) {
		HCT::DEBUG ("parser not defined.\n");
		$error = TRUE;
	}
	
	unless ( defined $self->{lexer} ) {
		HCT::DEBUG ("lexer not defined.\n");
		$error = TRUE;
	}
	
	return ($error == TRUE) ? FAILURE : SUCCESS;
}

#---------------------------------------------------------------------------
# Method:      parse
# Description: parse and close file.
# Returns:     result of yapp parser
#---------------------------------------------------------------------------

sub compute_metrics
{
	my $self   = shift;
	my $module = shift || $self->parser->module->{module};

	while ( $module ) {
		if ( defined $module->{inner} ) {
			$self->compute_metrics ($module);
		}

		my %metrics;

		while ( my ($name, $handler) = each %{ $module->{metrics} } ) {

			my %tmp_metrics;

			unless ( ref($handler) ) {
				$tmp_metrics{$name} = $metrics{$name} = $handler;
			}
			
			# get metrics produced by this metrics handler
			else {
				%tmp_metrics = $handler->compute( $self->output->{metrics} );
				while ( my ($k,$v) = each %tmp_metrics ) {
					$metrics{$k} = $v;
				}
			}

			# next metrics if we have none
			next unless scalar keys %tmp_metrics;

			if ( scalar(@{$self->{metrics}}) > 0 ) {
					while ( my ($metrics, $val) = each %tmp_metrics ) {
						if ( ! defined $self->{metrics}->[0]->[1]->{$metrics} ) {
							$self->{metrics}->[0]->[1]->{$metrics} = $val;
						}
						else {
							unless ( ref($handler) ) {
								$self->{metrics}->[0]->[1]->{$metrics} += $val;
							}
							else {
								$self->{metrics}->[0]->[1]->{$metrics} = 
								$handler->sum ( $self->{metrics}->[0]->[1]->{$metrics}, $val );
							}
						}
					}

			}
	
		}

		push @{ $self->{metrics} }, [
				$module->{name},
				{ %metrics } 
		];

		$module = $module->{next};
	}
}

sub parse
{
	my ($self, $filename) = (shift, shift);

	if ( $self->test != SUCCESS) {
		return FAILURE;
	}

	my $starttime = [ &Time::HiRes::gettimeofday() ];

	# FIXME: return?
	$self->parser->YYParse(
		yylex   => sub { $self->lexer->scan (@_) },
		yydebug => $HCT::YYDEBUG,
		yyerror => sub { },#sub { $self->parser->yyerror(@_) }
	);
	
	#print HCT::DUMP($self->parser->{metrics});
	$self->{metrics}->[0]->[1]->{TIME} = &Time::HiRes::tv_interval($starttime); 
}

=item
=cut

sub parser
{
	if (defined $_[1]) {
		$_[0]->{parser} = $_[1];
	}
	$_[0]->{parser};
}

sub lexer
{
	if (defined $_[1]) {
		$_[0]->{lexer} = $_[1];
		$_[0]->{lexer}->{handler}->{lang} = $_[0];
	}
	$_[0]->{lexer};
}

sub output
{
	$_[0]->{output};
}

sub input
{
	return shift->lexer->input (@_);
}

sub DESTROY
{
	shift->cleanup;	
}

sub cleanup
{

}

=back
=cut

1;


