# LexerDriver.pm (HCT::Lang::LexerDriver)
#
# Copyright 2008, 2009
# Slade Maurer
#
# This file is part of HCT, the Hardware Complexity Tool.
#
# HCT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# HCT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with HCT.  If not, see <http://www.gnu.org/licenses/>.

=head1 NAME

HCT::Language::LexerDriver - Lexer driver for languages under HCT system.

=head1 DESCRIPTION

Lexical analysis or scanning is the process where the stream of characters making up the
source program is read from left-to-right and grouped into tokens. Tokens are sequences
of characters with a collective meaning. There are usually only a small number of tokens
for a programming language: constants (integer, double, char, string, etc.), operators
(arithmetic, relational, logical), punctuation, and reserved words.

                       error messages 
                             |              
source language -> [ LEXICAL ANALYZER ] -> token stream

The lexical analyzer takes a source program as input, and produces a stream of tokens as
output. The lexical analyzer might recognize particular instances of tokens -- lexemes. 
A lexeme is the actual character sequence forming a token, the token is the general class that 
a lexeme belongs to. Some tokens have exactly one lexeme (e.g., the > character); for others, 
there are many lexemes (e.g., integer constants).

=cut

package HCT::Lang::LexerDriver;

our $VERSION = '0.80';

use HCT::Std;
use HCT::Std::IO;
use HCT::Constants;
use HCT::Lang::Token;
use HCT::Lang::LexerTracer;

use Switch;

use strict;
use warnings;

# Global subroutine AUTOLOAD to provide global reports.
sub AUTOLOAD
{
	return HCT::AUTOLOAD(
		(caller),
		our $AUTOLOAD,
		@_
	);
}

=head1 CONSTRUCTOR

=over 4

=item B<new ()>

Creates a new C<HCT::Lang::LexerDriver> object.

=back

=cut

sub new
{
	my $inv = shift;
	my $class = ref ($inv) || $inv;
	my $lexer = {
		input => undef,	# input handler

		@_,

		lines => [],  # set of passed lines
		line => "",   # current line content
		linenum	=> 0, # current line number
		linepos => 0, # current position

		tokens => {}, # hash of tokens: token name => token object
		patterns => [], # list of patterns
		stack => [], # stack of tokens

		parser => undef, # parser handler
		tracer => undef, # tracer handler
	};

	bless $lexer, $class;

	# Initialize tracer for lexer (LexerTracer)
	$lexer->{tracer} = HCT::Lang::LexerTracer->new (
		line => \$lexer->{'line'},
		linepos	=> \$lexer->{'linepos'},
		linenum	=> \$lexer->{'linenum'},
		stack => \$lexer->{'stack'}
	);

	# Do token reservation
	$lexer->do_reserve;

	# Return new lexer object
	return $lexer;
}

=head1 DESTRUCTOR

=over 4

=item B<DESTROY ()>

Destroy C<HCT::Lang::LexerDriver> object.

=back

=cut

sub DESTROY
{
	my $lexer = shift;

	# If defined input handler for lexer, then we
	# have to destroy it
	if (defined $lexer->input) {
		# Close input handler
		$lexer->input->close;

		# Destroy input handler 
		undef ($lexer->{input});
	}
}

=head1 METHODS
=cut

=over 4
=cut

=item B<input ([FH])>

Returns input object created by C<HCT::Std::IO>. If FH set, 
creates a new input object. 

=cut

sub input
{
	my ($lexer, $fh) = (shift, shift);

	# If file handler was already set and we do not
	# need change it, return it
	return $lexer->{input} unless defined $fh;

	# Initialize lexer input
	$lexer->{input} = new HCT::Std::IO::Handle;

	# If handler was created successful
	if ($fh && ref ($fh) eq 'GLOB') {
		unless ($lexer->input->fdopen (fileno ($fh), 'r')) {
			HCT::ERROR ("Cannot work with such input handler: ". ref ($fh) .".\n");
		}
	} else {
		HCT::ERROR ("Bad input handler: " . ref ($fh) . ".\n")
	}
}

=item B<skip_whitespace ()>

Skips whitespaces. Just removes leading white spaces.

=cut

sub skip_whitespace
{
	$_[0]->match (qr/\s*/);
}

=item B<match (EXPR)>

If EXPR passed successfully, returns matched value and sets 
new line position. Else, goes back to the previous 
position and returns C<undef>.

=cut

sub match
{
	my ($self, $expr) = (shift, shift);

	if ($self->{line} =~ /\G$expr/g) {
		my $match = substr ($self->{line}, $self->linepos, pos ($self->{line}) - $self->linepos);
		$self->linepos (pos ($self->{line}));
		return $match;
	}
	
	# Put the position back
	pos ($self->{line}) = $self->linepos;
	# and return undefined value
	return undef;
}

=item B<linepos ([POS])>

If POS set, changes current line position. Returns current 
line position.

Important: works not directly with current line position (C<pos>), 
but with variable that store previous position value.

=cut

sub linepos
{
	my ($self, $pos) = (shift, shift);
	$self->{linepos} = $pos
		if defined $pos; 
	return $self->{linepos};
}

=item B<linelen ()>

Returns current line length.

=cut

sub linelen
{
	my $len = length ($_[0]->{line}) || 0;
	return $len;
}

=item B<linenum ()>

Returns current line number.

=cut

sub linenum
{
	return $_[0]->{linenum};
}

=item B<is_eof ()>

Returns true if end of file (input), or false.

=cut

sub is_eof
{
	my $lexer = shift;
	return ($lexer->input->eof) ? TRUE : FALSE;
}

=item B<is_eol ()> 

Returns true if end of line, or false.

=cut

sub is_eol
{
	my $lexer = shift;
	return ($lexer->linepos == $lexer->linelen) ? TRUE : FALSE;
}

=item B<move ()>

Returns can we moved forward or no. If line is not ended (C<pos> equal to C<len>), 
returns C<TRUE>, else gets a new line while line is empty. If file is finished returns C<FALSE>.

=cut

sub move
{
	my $lexer = shift;

	return TRUE if !$lexer->is_eol;
	
	$lexer->event_eol if $lexer->linenum > 0;

	if ($lexer->get_next_line) {
		while ($lexer->is_emptyline) {
			$lexer->get_next_line;

			# if we have the end of file
			return FALSE if $lexer->is_eof && $lexer->is_eol;
		}

		return TRUE;
	}

	return FALSE;
}

=item B<is_emptyline ()>

Returns C<TRUE> if current line is empty.

=cut

sub is_emptyline
{
	my $lexer = shift;
	if ($lexer->{'line'} =~ /^\s*$/) {
		$lexer->match (qr/\s*$/);
		return TRUE;
	}
	return FALSE; 
}

=item B<get_next_line ()>

Gets a new line and returns C<TRUE> if the new line has been received. 
Or C<FALSE> if C<EOF>.

=cut

sub get_next_line
{
	my $lexer = shift;
	return FALSE if $lexer->is_eof;
	
	# save current line
	$lexer->{lines}->[$lexer->linenum] = $lexer->{line}
		if $lexer->linenum > 0;

	# get a new line
	$lexer->{line} = $lexer->input->getline;
	$lexer->{linenum} += 1;

	# set current line position to 0
	$lexer->linepos (0);
	
	return TRUE;
}

=item B<scan (PARSER)>

The scanner has encoded within it information on the possible 
sequences of characters that can be contained within any of 
the tokens it handles. Returns token.

=cut

sub scan
{
	my $lexer;

	$lexer = shift;
	$lexer->{parser} = shift;

	# Infinite loop while we're getting the next token
	LOOP: while (TRUE) {

		# Try to move ahead. If we cannot do this, then 
		# lexer have to stop.
		unless ($lexer->move) {
			$lexer->stop;
			return $lexer->shift_token;
		}

		# If current position is in the comment or there is a comment, 
		# we will skip this space
		next LOOP if $lexer->skip_comment;

		# Skip white space from the current state
		$lexer->skip_whitespace;

		# Hook to skip a special momments of the language grammar
		next LOOP if $lexer->skip_this;

		# Be careful, if line is empty then skip it and rescan input
		next LOOP if $lexer->is_eol;

		# Try to get next token. If true, than return it last token
		# from the stack
		return $lexer->shift_token if $lexer->next_token;

		# Appeared syntax error or maybe corresponding 
		# action was not defined.
		$lexer->tracer->error ("Syntax error.\n");
		return EMPTY_TOKEN;
	}
	
	# We can not be here!
	HCT::ERROR ();
}

=item B<stop ()>

Returns C<EMPTY_TOKEN> to stop the parser. Could be rewritten if needed, see CDL lexer.

=cut

sub stop
{
	$_[0]->push_token (EMPTY_TOKEN);
}

=item B<callback ()>

Returns parser handler. Will be updated as soon, as C<scan> called.

=cut

sub callback
{
	shift->{parser};
}

=item B<tracer ()>

Returns object of tracing created by C<L<HCT::Lang::LexerTracer>>.

=cut

sub tracer
{
	shift->{tracer};
}

=item B<next_token ()>

Tries to find corresponding token. Return true if token was 
found, or false as finish.

=cut

sub next_token
{
	my $lexer = shift;

	$lexer->tracer->debug if $HCT::YYDEBUG;

	return TRUE if $lexer->checkup_available_tokens;

	return TRUE if $lexer->match_pattern ($lexer->{'patterns'});

	return FALSE;	
}

=item B<match_pattern ()>

=cut

sub match_pattern
{
	my $lexer = shift;
	my $patterns = shift;
	
	foreach my $pattern (@$patterns) {
		if (defined (my $lexeme = $lexer->match ($pattern->{'pattern'}))) {
			my $token = $lexer->get_token ($pattern->{'name'});
			return $lexer->push_token ($token->get_name, $lexeme);
		}
	}

	return FALSE;
}

=item B<checkup_available_tokens ()>

=cut

sub checkup_available_tokens
{
	my $lexer = shift;

	# get last state number
	my $last_state_num = $lexer->callback->{STACK}->[-1]->[0];
	
	# get current parser state
	my $state = $lexer->callback->{STATES}->[$last_state_num];

	# get posible parser actions
	my $actions = $state->{ACTIONS};

	# get default state
	my $default = $state->{DEFAULT};

	# if we have to choose between predefined list of actions
	if (scalar (keys %$actions) > 0 && !defined $default) {
		my @patterns;
		map {
			push @patterns, $_ if defined $actions->{$_->{'name'}} 
		} @{$lexer->{'patterns'}};

		return TRUE if $lexer->match_pattern (\@patterns);
	}
	return FALSE;
}

=item B<do_reserve ()>

Makes reservation of tokens by calling C<reserve_tokens>.

=cut

sub do_reserve
{
	my $lexer = shift;

	# Reserve new tokens
	$lexer->{'tokens'} = { };
	$lexer->reserve_tokens;
	
	$lexer->{'patterns'} = [ ];
	$lexer->reserve_patterns;
}

=item B<reserve_patterns ()>

Creates a list of patterns in C<patterns>. Each element has such fields: tokens name, type and 
one pattern. That means, we create a sigle element for each new pattern. After creating this 
list will be sorted by types and patterns.

=cut

sub reserve_patterns
{
	my $lexer = shift;

	while (my ($name, $token) = each %{$lexer->{'tokens'}}) {
		foreach my $pattern ($token->get_patterns) {
			push @{$lexer->{'patterns'}}, {
				'name' => $token->get_name,
				'type' => $token->get_type,
				'pattern' => $pattern,
			};
		}
	}

	@{$lexer->{'patterns'}} = sort {
		
		return ($b->{'pattern'} cmp $a->{'pattern'}) if ($a->{'type'} == $b->{'type'});
		return (-1) if ($a->{'type'} > $b->{'type'});
		return (+1) if ($a->{'type'} < $b->{'type'});

	} @{$lexer->{'patterns'}};
}

=item B<reserve (TYPE, TOKENS)>

Creates an objects of new tokens by using C<L<HCT::Lang::Token>> and 
stores them into the C<tokens>.

=cut

sub reserve
{
	my $lexer = shift;
	my ($type, @tokens_descriptions) = (shift, @_);

	if ($type !~ /^[\d]+$/) {
		HCT::DEBUG ("The token's type may consists only of numbers.\n");
		return;
	}

	TOKEN: foreach my $description (@tokens_descriptions) {
		my ($name, $patterns) = @$description;

		# Error if token with such name was already defined
		if (defined $lexer->{'tokens'}->{$name}) {
			HCT::DEBUG ("Token '$name' was already defined. Skipped.\n");
			next TOKEN;
		}

		# Create new token object. Move to the next token if object is undefined
		my $token = HCT::Lang::Token->new (
			'name' => $name,
			'type' => $type,
			'patterns' => $patterns,
		);

		next TOKEN if !defined $token;
		$lexer->{'tokens'}->{$name} = $token;
	}
}

=item B<get_token (NAME)>

Gets token name and returns token object from C<tokens>.

=cut

sub get_token
{
	my ($lexer, $name) = (shift, shift);
	return $lexer->{'tokens'}->{$name};
}

=item B<shift_token ()>

Shifts and returns token from token stack.

=cut

sub shift_token
{
	my $lexer = shift;
	my $token = shift @{ $lexer->{tokens}->{stack} };
	return @$token;
}

=item B<push_token ()>

Psuh new token to the stack and returns true.

=cut

sub push_token
{
	my $lexer = shift;
	my @token = @_;
	push @{ $lexer->{tokens}->{stack} }, \@token;
	return TRUE;
}

=back
=cut

=head1 VIRTUAL METHODS
=cut

=over 4
=cut

=item B<skip_comment ()>

Virtual method that provides skip from the comments. Returns 
true if current position should be skipped.

=cut

sub skip_comment
{
	return FALSE;
}
=item B<skip_this ()>

Virtual method that provides skip some stuff from the current positions. 
Returns true or false.

=cut

sub skip_this
{
	return FALSE;
}

=item B<reserve_tokens ()> 

Virtual method to provide reserve of tokens.

=cut

sub reserve_tokens
{
	HCT::DEBUG ("method 'reserve_tokens' was not replaced.");
}

=back
=cut

=head1 PROBLEMS

=head2 Identifiers and keywords

The lexer state solution involves coordination between the lexer and
the parser. In particular, the parser must tell the lexer whether in
"this context" a keyword is expected or an identifier is
expected. There are several problems with coordinating the lexer's
state with the parser's context. One of the most frequently noted
ones is that it makes (multiple token) lookahead more difficult. A
related problem occurs if the context is miscommunicated between the
parser and the lexer, the lexer may return a keyword when only
identifers are expected or return the keyword as an identifier when
the keyword was supposed to be treated as a keyword.

=cut

1;

