# Lexer.pm - Verilog Lexer
#
# Copyright 2008, 2009
# Slade Maurer
#
# This file is part of HCT, the Hardware Complexity Tool
#
# HCT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# HCT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with HCT.  If not, see <http://www.gnu.org/licenses/>.

package HCT::lang::hdl::verilog::lexer;

use base 'HCT::Lang::LexerDriver';

use HCT::Constants;
use HCT::Lang::Token;

use strict;
use warnings;

use constant {
	operator => 4,
	number => 3,
	keyword => 2,
	identifier => 1,
	misc => 0,
};

# Reserve tokens for lexer.

sub reserve_tokens
{
	my $lexer = shift;

	# OPERATORS

	$lexer->reserve (
		operator,

		['YYRIGHTARROW',     ['->']],
		['YYALLPATH',        ['*>']],
		['YYNBASSIGN',       ['<=']],
		['YYLEADTO',         ['=>']],
		['YYGEQ',            ['>=']],
		['YYLEQ',            ['=<']],
		['YYLOGAND',         ['&&']],
		['YYLOGOR',          ['||']],
		['YYCASEEQUALITY',   ['===']],
		['YYLOGEQUALITY',    ['==']],
		['YYCASEINEQUALITY', ['!==']],
		['YYLOGINEQUALITY',  ['!=']],
		['YYLOGXNOR',        ['~^','^~']],
		['YYLOGNAND',        ['~&']],
		['YYLOGNOR',         ['~|']],
		['YYLSHIFT',         ['<<']],
		['YYRSHIFT',         ['>>']],
		['YYCONDITIONAL',    ['?:']],
	);

	# IDENTIFIERS

	$lexer->reserve (
		identifier,

		['YYID',[qr/\`[a-zA-Z_][a-zA-Z0-9_]*/,qr/[a-zA-Z_][a-zA-Z0-9_]*/]],
	);

	# NUMBERS

	$lexer->reserve (
		number,

		['YYINUMBER',[qr/(?:[0-9][0-9_]*)\'[hH][ ]*[0-9a-fA-FxXzZ?][0-9a-fA-FxXzZ?_]*/,
		              qr/(?:[0-9][0-9_]*)\'[bB][ ]*[01xXzZ?][01xXzZ?_]*/,
		              qr/[0-9][0-9_]*(?!['hH|0-9])/]],
		['YYRNUMBER',[qr/([0-9][0-9_]*)*\.([0-9][0-9_]*)+/,
		              qr/([0-9][0-9_]*)+\.([0-9][0-9_]*)*/]],
	);

	# KEYWORDS

	$lexer->reserve (
		keyword,

		['YYsysID',        [qr/\$[a-zA-Z][a-zA-Z0-9]+/]],
		['YYsysSETUP',     ['$setup']],
		['YYsysND',        ['$NDSET','$ND']],
		['YYSTRING',       [qr/\"[^"]*\"/]],
		['YYXNOR',         ['xnor']],
		['YYXOR',          ['xor']],
		['YYWOR',          ['wor']],
		['YYVECTORED',     ['vectored']],
		['YYWAIT',         ['wait']],
		['YYWAND',         ['wand']],
		['YYWEAK0',        ['weak0']],
		['YYWEAK1',        ['weak1']],
		['YYTESLATIMER',   ['teslaTimer']],
		['YYTIME',         ['time']],
		['YYTRAN',         ['tran']],
		['YYTRANIF0',      ['tranif0']],
		['YYTRANIF1',      ['tranif1']],
		['YYSTRONG0',      ['strong0']],
		['YYSTRONG1',      ['strong1']],
		['YYSUPPLY0',      ['supply0']],
		['YYSUPPLY1',      ['supply1']],
		['YYTABLE',        ['table']],
		['YYRNMOS',        ['rnmos']],
		['YYRPMOS',        ['rpmos']],
		['YYRTRAN',        ['rtran']],
		['YYRTRANIF0',     ['rtranif0']],
		['YYRTRANIF1',     ['rtranif1']],
		['YYSCALERED',     ['scalered']],
		['YYSMALL',        ['small']],
		['YYSPECPARAM',    ['specparam']],
		['YYRCMOS',        ['rcmos']],
		['YYREAL',         ['real']],
		['YYREPEAT',       ['repeat']],
		['YYNOT',          ['not']],
		['YYNOTIF0',       ['notif0']],
		['YYNOTIF1',       ['notif1']],
		['YYOR',           ['or']],
		['YYPMOS',         ['pmos']],
		['YYPULL0',        ['pull0']],
		['YYPULL1',        ['pull1']],
		['YYPULLDOWN',     ['pulldown']],
		['YYPULLUP',       ['pullup']],
		['YYDEFAULT',      ['default']],
		['YYDEFPARAM',     ['defparam']],
		['YYDISABLE',      ['disable']],
		['YYEDGE',         ['edge']],
		['YYFUNCTION',     ['function']],
		['YYHIGHZ0',       ['highz0']],
		['YYHIGHZ1',       ['highz1']],
		['YYENDFUNCTION',  ['endfunction']],
		['YYENDPRIMITIVE', ['endprimitive']],
		['YYSPECIFY',      ['specify']],
		['YYENDSPECIFY',   ['endspecify']],
		['YYENDTABLE',     ['endtable']],
		['YYTASK',         ['task']],
		['YYENDTASK',      ['endtask']],
		['YYENUM',         ['enum']],
		['YYEVENT',        ['event']],
		['YYCMOS',         ['cmos']],
		['YYAND',          ['and']],
		['YYINOUT',        ['inout']],
		['YYBUFIF0',       ['bufif0']],
		['YYBUFIF1',       ['bufif1']],
		['YYMACROMODULE',  ['macromodule']],
		['YYMODULE',       ['module']],
		['YYENDMODULE',    ['endmodule']],
		['YYFORK',         ['fork']],
		['YYFOR',          ['for']],
		['YYFOREVER',      ['forever']],
		['YYIF',           ['if']],
		['YYELSE',         ['else']],
		['YYBUF',          ['buf']],
		['YYINITIAL',      ['initial']],
		['YYALWAYS',       ['always']],
		['YYDEASSIGN',     ['deassign']],
		['YYASSIGN',       ['assign']],
		['YYCASE',         ['case']],
		['YYCASEX',        ['casex']],
		['YYCASEZ',        ['casez']],
		['YYPOSEDGE',      ['posedge']],
		['YYPRIMITIVE',    ['primitive']],
		['YYBEGIN',        ['begin']],
		['YYENDCASE',      ['endcase']],
		['YYEND',          ['end']],
		['YYTYPEDEF',      ['typedef']],
		['YYOUTPUT',       ['output']],
		['YYREG',          ['reg']],
		['YYINPUT',        ['input']],
		['YYINTEGER',      ['integer']],
		['YYJOIN',         ['join']],
		['YYLARGE',        ['large']],
		['YYMEDIUM',       ['medium']],
		['YYMREG',         ['mREG']],
		['YYNAND',         ['nand']],
		['YYNEGEDGE',      ['negedge']],
		['YYNMOS',         ['nmos']],
		['YYNOR',          ['nor']],
		['YYSWIRE',        ['swire']],
		['YYWIRE',         ['wire']],
		['YYTRI0',         ['tri0']],
		['YYTRI1',         ['tri1']],
		['YYTRIAND',       ['triand']],
		['YYTRIOR',        ['trior']],
		['YYTRIREG',       ['trireg']],
		['YYTRI',          ['tri']],
		['YYPARAMETER',    ['parameter']],
	);

	# MISC

	$lexer->reserve (
		misc,
		
		[';',[';']],
		['(',['(']],
		[')',[')']],
		['[',['[']],
		[']',[']']],
		['=',['=']],
		['@',['@']],
		['#',['#']],
		[',',[',']],
		[':',[':']],
		['.',['.']],
		['<',['<']],
		['+',['+']],
		['*',['*']],
		['-',['-']],
		['/',['/']],
		['{',['{']],
		['}',['}']],
		['?',['?']],
	);

}

sub event_eol
{
	my $lexer = shift;
	
	if (defined $lexer->callback->module->{current}) {
		$lexer->callback->module->{current}->{metrics}->{SLOC} += 1;
	} else {
		$lexer->{handler}->{lang}->{metrics}->[0]->[1]->{'SLOC'} += 1; 
	}
}

sub skip_comment
{
	my $lexer = shift;
	my $multiline_comment = FALSE;

	# linecomment
	if (defined $lexer->match (qr/\s*\/\//)) {
		
		if (defined $lexer->callback->module->{current}) {
			$lexer->callback->module->{current}->{metrics}->{COMMENT_LINES} += 1;
		} else {
			$lexer->{handler}->{lang}->{metrics}->[0]->[1]->{COMMENT_LINES} += 1; 
		}
		
		
		$lexer->linepos ($lexer->linelen);
		return TRUE
	}

	# multiline comment
	if (defined $lexer->match(qr/\s*\/\*/) ) {
		$multiline_comment = TRUE;
	}
	
	while ($multiline_comment == TRUE) {

		if (defined $lexer->callback->module->{current}) {
			$lexer->callback->module->{current}->{metrics}->{COMMENT_LINES} += 1;
		} else {
			$lexer->{handler}->{lang}->{metrics}->[0]->[1]->{COMMENT_LINES} += 1; 
		}

		if (defined $lexer->match (qr/(.*?)\*\//) ) {
			return TRUE;
		}
		$lexer->linepos ($lexer->linelen);
		$lexer->move;
	}

	return FALSE;
}

sub skip_this
{
	my $lexer = shift;
	
	# Macros:
	return TRUE if defined $lexer->match ('\`timescale[^\n]*\n');
	return TRUE if defined $lexer->match ('\`include[^\n]*\n');

	if (   defined $lexer->match ('\`define[^\n]*\n')
		|| defined $lexer->match ('\`ifdef[^\n]*\n')
		|| defined $lexer->match ('\`ifndef[^\n]*\n')
		|| defined $lexer->match ('\`else')
		|| defined $lexer->match ('\`endif')
	  ) {
		return TRUE;
	}

	return FALSE;
}

1; 


